-- Reference design - Frequency counter for Spartan-3E Starter Kit (Rev.C).
--
-- Ken Chapman - Xilinx Ltd - 7th March 2006
--
--  *** This design contains an evaluation test feature of the DCM. ****
--  *** Before this design can be processed a special environment   ****
--  *** needs to be set or the 'dcm_fixed_osc' module removed.      ****
--  *** Please read the notes provided in 'dcm_fixed_osc.vhd' for   ****
--  *** details of this special requirement.                        ****
--
--
-- Provides a frequency counter with resolution of 1Hz up to a maximum frequency of 
-- approximately 200MHz (limited by maximum clock rate of design).
--
-- The 50MHz on board oscillator is used to provide a 1 second time base and therefore 
-- defines the accuracy of the measurement. If the frequency of the 50MHz oscillator can 
-- be determined using a calibrated frequency counter or frequency generator then this design 
-- could be tuned to match the particular board. 
--
-- Slide switches SW0, SW1, SW2 and SW3 allow selection of the input signal or one of three test 
-- signals. Only one switch should be selected and the LCD display will guide and confirm the 
-- selection process.
--
-- SW0 selects SMA connector (J17) as the input signal to be analysed. 
-- SW1 selects the 50MHz oscillator as a test signal and must yield a perfect result since it is 
--     also the definition of the measurement time base..
-- SW2 selects a signal of approximately 145MHz which is generated by one of the DCMs operating 
--     in a special test oscillator mode.
-- SW3 selects a ring oscillator implement using a combinatorial loop in the CLB fabric. 
--
-- LED indicate measurements in progress (1 second alternate flashing of 4 LEDs).
--
-- PicoBlaze performs calculations and conversions as well as driving the LCD display.
--
------------------------------------------------------------------------------------
--
-- NOTICE:
--
-- Copyright Xilinx, Inc. 2006.   This code may be contain portions patented by other 
-- third parties.  By providing this core as one possible implementation of a standard,
-- Xilinx is making no representation that the provided implementation of this standard 
-- is free from any claims of infringement by any third party.  Xilinx expressly 
-- disclaims any warranty with respect to the adequacy of the implementation, including 
-- but not limited to any warranty or representation that the implementation is free 
-- from claims of any third party.  Furthermore, Xilinx is providing this core as a 
-- courtesy to you and suggests that you contact all third parties to obtain the 
-- necessary rights to use this implementation.
--
------------------------------------------------------------------------------------
--
-- Library declarations
--
-- Standard IEEE libraries
--
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;
--
-- The Unisim Library is used to define Xilinx primitives. It is also used during
-- simulation. The source can be viewed at %XILINX%\vhdl\src\unisims\unisim_VCOMP.vhd
--
library unisim;
use unisim.vcomponents.all;
--
------------------------------------------------------------------------------------
--
--
entity frequency_counter is
    Port (             led : out std_logic_vector(7 downto 0);
                        sw : in std_logic_vector(3 downto 0);
            strataflash_oe : out std_logic;
            strataflash_ce : out std_logic;
            strataflash_we : out std_logic;
                     lcd_d : inout std_logic_vector(7 downto 4);
                    lcd_rs : out std_logic;
                    lcd_rw : out std_logic;
                     lcd_e : out std_logic;
                   sma_clk : in std_logic;
                 clk_50mhz : in std_logic);
    end frequency_counter;
--
------------------------------------------------------------------------------------
--
-- Start of test architecture
--
architecture Behavioral of frequency_counter is
--
------------------------------------------------------------------------------------
--
-- declaration of KCPSM3
--
  component kcpsm3 
    Port (      address : out std_logic_vector(9 downto 0);
            instruction : in std_logic_vector(17 downto 0);
                port_id : out std_logic_vector(7 downto 0);
           write_strobe : out std_logic;
               out_port : out std_logic_vector(7 downto 0);
            read_strobe : out std_logic;
                in_port : in std_logic_vector(7 downto 0);
              interrupt : in std_logic;
          interrupt_ack : out std_logic;
                  reset : in std_logic;
                    clk : in std_logic);
    end component;
--
-- declaration of program ROM used by data capture processor
--
  component fc_ctrl
    Port (      address : in std_logic_vector(9 downto 0);
            instruction : out std_logic_vector(17 downto 0);
             proc_reset : out std_logic;                       --JTAG Loader version
                    clk : in std_logic);
    end component;

--
-- Fixed frequency oscillator using a DCM
--
component dcm_fixed_osc 
  port(    clk_out : out std_logic;
        kick_start : in std_logic   );
  end component;  

--
-- Ring oscillator
--
component ring_osc
  port(   osc_out : out std_logic;
            reset : in std_logic   );
   end component;  

--
------------------------------------------------------------------------------------
--
-- Signals used to select and provide sources to be analysed
--
signal source_control          : std_logic_vector(7 downto 0);
signal dcm_oscillator          : std_logic;
signal ring_oscillator         : std_logic;
signal freq_for_measurement    : std_logic;
signal test_clk                : std_logic;
--
--
--
--
-- Signals used by clock cycle counters and controls
--
signal ab_switch_delay         : std_logic_vector (3 downto 0);
signal a_count_ce              : std_logic;
signal a_count_rst             : std_logic;
signal a_count                 : std_logic_vector (31 downto 0);
signal b_count_ce              : std_logic;
signal b_count_rst             : std_logic;
signal b_count                 : std_logic_vector (31 downto 0);
--
--
--
-- Signals for 1 second interrupt generation and counter switching
--
signal one_second_count       : integer range 0 to 49999999 :=0;
signal one_second_pulse       : std_logic;
signal ab_switch              : std_logic;
signal interrupt_delay        : std_logic_vector (99 downto 0);
--
--
-- Signals used to connect KCPSM3 to program ROM and I/O logic
--
signal address        : std_logic_vector(9 downto 0);
signal instruction    : std_logic_vector(17 downto 0);
signal port_id        : std_logic_vector(7 downto 0);
signal out_port       : std_logic_vector(7 downto 0);
signal in_port        : std_logic_vector(7 downto 0);
signal write_strobe   : std_logic;
signal read_strobe    : std_logic;
signal interrupt      : std_logic;
signal interrupt_ack  : std_logic;
signal reset          : std_logic;
--
--
-- Signals for LCD operation
--
-- Tri-state output requires internal signals
-- 'lcd_drive' is used to differentiate between LCD and StrataFLASH communications 
-- which share the same data bits.
--
signal   lcd_rw_control : std_logic;
signal  lcd_output_data : std_logic_vector(7 downto 4);
signal        lcd_drive : std_logic;
--
--
--
------------------------------------------------------------------------------------------------------------------------------------------------------------------------
--
-- Start of circuit description
--
begin
  --
  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- Disable unused components  
  ----------------------------------------------------------------------------------------------------------------------------------
  --
  --StrataFLASH must be disabled to prevent it conflicting with the LCD display 
  --
  strataflash_oe <= '1';
  strataflash_ce <= '1';
  strataflash_we <= '1';
  --
  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- Select and Buffer signal to be analysed
  ----------------------------------------------------------------------------------------------------------------------------------
  --
  -- 
  -- The fixed frequency dcm oscillator using test mode of the DCM
  --
  dcm_fixed_oscillator: dcm_fixed_osc
  port map (    clk_out => dcm_oscillator,
             kick_start => source_control(7) );

  -- 
  -- ring Oscillator using logic fabric 
  --

  logic_oscillator: ring_osc
  port map (   osc_out => ring_oscillator,
                 reset => source_control(6) );

  -- 
  -- Source selection multiplexer (must be combinatorial) 
  --

  freq_for_measurement <= sma_clk when (source_control(1 downto 0)="00") 
                     else clk_50mhz when (source_control(1 downto 0)="01") 
                     else dcm_oscillator when (source_control(1 downto 0)="10") 
                     else ring_oscillator;


  -- 
  -- Global buffer on selected source 
  --

  buffer_clkin: BUFG
  port map( O => test_clk,
            I => freq_for_measurement);

  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- Frequency Counters
  --
  -- This is formed of two simple 32-bit binary counters such that one can be counting whilst the 
  -- other is being read and reset. Great care is taken to ensure the switch over between the two 
  -- counters is clean and that no clock cycles are missed or glitches generated.
  --
  -- An asynchronous reset is provided and will be controlled by PicoBlaze using a long
  -- pulse when the counter is disabled. Hence there is no special synchronisation on this control.
  -- 
  ----------------------------------------------------------------------------------------------------------------------------------
  --

  counter_switch_control: process(test_clk)
  begin
    if test_clk'event and test_clk='1' then

      -- Four stage register to ensure synhchronisation of control between clock domains
      ab_switch_delay <= ab_switch_delay(2 downto 0) & ab_switch;
      
      -- Separate enables form switch between counters when control is consistantly high or low.

      case ab_switch_delay(3 downto 1) is

        when "000" => a_count_ce <= '1';
                      b_count_ce <= '0';

        when "111" => a_count_ce <= '0';
                      b_count_ce <= '1';

        when others => a_count_ce <= a_count_ce;
                       b_count_ce <= b_count_ce;

      end case;

    end if;
  end process counter_switch_control;

  -- 32-bit counters 

  test_counter_a: process(test_clk, a_count_rst )
  begin
    if a_count_rst='1' then
       a_count <= X"00000000";
     elsif test_clk'event and test_clk='1' then
      if a_count_ce='1' then
         a_count <= a_count + 1;
       else
         a_count <= a_count;
      end if;
    end if;
  end process test_counter_a;

  test_counter_b: process(test_clk, b_count_rst )
  begin
    if b_count_rst='1' then
       b_count <= X"00000000";
     elsif test_clk'event and test_clk='1' then
      if b_count_ce='1' then
         b_count <= b_count + 1;
       else
         b_count <= b_count;
      end if;
    end if;
  end process test_counter_b;


  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- One Second Timer and Circuit Control
  --
  -- Using the normal 50MHz clock source, this circuit generates a 1 second timer which is used to  
  -- switch the enable from one counter to the other. 
  --
  -- A short delay is then inserted to ensure that the switch has taken place before interrupting the 
  -- PicoBlaze processor. This ensures that the counters are switched and stable for reading etc.
  -- Note: This could cause an issue when very low frequency signals are being measured.
  -- 
  ----------------------------------------------------------------------------------------------------------------------------------
  --
  one_second_logic: process(clk_50mhz)
  begin
    if clk_50mhz'event and clk_50mhz='1' then

      if one_second_count=49999999 then      --divide by 50,000,000 is 1s
        one_second_count <= 0;
        one_second_pulse <= '1';
       else
        one_second_count <= one_second_count + 1;
        one_second_pulse <= '0';
      end if;

      -- delay of 100 clock cycles before generating interrupt 
      interrupt_delay <= interrupt_delay(98 downto 0) & one_second_pulse;

      -- processor interrupt waits for an acknowledgement
      if interrupt_ack='1' then
         interrupt <= '0';
        elsif interrupt_delay(99) ='1' then
         interrupt <= '1';
        else
         interrupt <= interrupt;
      end if;

      -- counter selection switch toggles each second
      if one_second_pulse='1' then
        ab_switch <= not ab_switch;
      end if;

    end if;

  end process one_second_logic;

  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- KCPSM3 and the program memory used for capturing data from the counters
  ----------------------------------------------------------------------------------------------------------------------------------
  --

  processor: kcpsm3
    port map(      address => address,
               instruction => instruction,
                   port_id => port_id,
              write_strobe => write_strobe,
                  out_port => out_port,
               read_strobe => read_strobe,
                   in_port => in_port,
                 interrupt => interrupt,
             interrupt_ack => interrupt_ack,
                     reset => reset,
                       clk => clk_50mhz);
 
  program_rom: fc_ctrl
    port map(      address => address,
               instruction => instruction,
                proc_reset => reset,        --JTAG Loader version
                       clk => clk_50mhz);

  --reset <= '0';  --When using normal version
  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- Capture processor input ports 
  ----------------------------------------------------------------------------------------------------------------------------------
  --
  --
  -- The inputs are connected via a pipelined multiplexer
  --

  input_ports: process(clk_50mhz)
  begin
    if clk_50mhz'event and clk_50mhz='1' then

      case port_id(7 downto 4) is

        -- read A-counter for addresses 00, 10, 20 and 30 hex
        when "0000" =>    in_port <= a_count(7 downto 0);  
        when "0001" =>    in_port <= a_count(15 downto 8);  
        when "0010" =>    in_port <= a_count(23 downto 16);  
        when "0011" =>    in_port <= a_count(31 downto 24);  

        -- read B-counter for addresses 40, 50, 60 and 70 hex
        when "0100" =>    in_port <= b_count(7 downto 0);  
        when "0101" =>    in_port <= b_count(15 downto 8);  
        when "0110" =>    in_port <= b_count(23 downto 16);  
        when "0111" =>    in_port <= b_count(31 downto 24);  

        -- read slide switches and counter circuit status at address 80 hex
        when "1000" =>    in_port <= "000" & ab_switch & sw;

        -- read LCD data at address 90 hex
        when "1001" =>    in_port <= lcd_d & "0000";

        -- Don't care used to ensure minimum logic
        when others =>    in_port <= "XXXXXXXX";  

      end case;

    end if;

  end process input_ports;


  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- Capture processor output ports 
  ----------------------------------------------------------------------------------------------------------------------------------
  --

  output_ports: process(clk_50mhz)
  begin

    if clk_50mhz'event and clk_50mhz='1' then
      if write_strobe='1' then
        
        -- LED register at address 01 hex 
        if port_id(0)='1' then
          led <= out_port;
        end if;

        -- Counter reset controls at address 02 hex 
        if port_id(1)='1' then
          a_count_rst <= out_port(0);
          b_count_rst <= out_port(1);
        end if;

        -- LCD data output and controls at address 04 hex.

        if port_id(2)='1' then
          lcd_output_data <= out_port(7 downto 4);
          lcd_drive <= out_port(3);  
          lcd_rs <= out_port(2);
          lcd_rw_control <= out_port(1);
          lcd_e <= out_port(0);
        end if;

        -- Source selection and control at address 08 hex.

        if port_id(3)='1' then
          source_control <= out_port;
        end if;

      end if;
    end if; 

  end process output_ports;

  --
  --
  ----------------------------------------------------------------------------------------------------------------------------------
  -- LCD interface  
  ----------------------------------------------------------------------------------------------------------------------------------
  --
  -- The 4-bit data port is bidirectional.
  -- lcd_rw is '1' for read and '0' for write 
  -- lcd_drive is like a master enable signal which prevents either the 
  -- FPGA outputs or the LCD display driving the data lines.
  --
  --Control of read and write signal
  lcd_rw <= lcd_rw_control and lcd_drive;

  --use read/write control to enable output buffers.
  lcd_d <= lcd_output_data when (lcd_rw_control='0' and lcd_drive='1') else "ZZZZ";

  ----------------------------------------------------------------------------------------------------------------------------------


------------------------------------------------------------------------------------------------------------------------------------

end Behavioral;

------------------------------------------------------------------------------------------------------------------------------------
--
-- END OF FILE frequency_counter.vhd
--
------------------------------------------------------------------------------------------------------------------------------------

